package org.expeditee.agents;

import java.util.HashMap;
import java.util.LinkedList;

import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameIO;
import org.expeditee.gui.FrameUtils;
import org.expeditee.items.Item;

public class CopyTree extends TreeProcessor {

	private String _nameTo;

	private String _nameFrom;

	private String _framePath;

	private int _lastNumber = -1;

	private int _firstNumber = 1;

	private HashMap<String, String> _nameMap = new HashMap<String, String>();

	private LinkedList<String> _toReparse = new LinkedList<String>();

	public CopyTree(String framesetTo) {
		_nameTo = framesetTo.trim();
	}

	@Override
	public boolean initialise(Frame init) {
		_nameFrom = init.getFramesetName().toLowerCase();

		// get the last used frame in the destination frameset
		_lastNumber = FrameIO.getLastNumber(_nameTo);

		if (_lastNumber < 0) {
			// create the new frameset
			Frame one = FrameIO.CreateFrameset(_nameTo, init.path);
			_framePath = one.path;
			_lastNumber = -1;
			_firstNumber = 1;

			// copy the original .0 frame
			Frame zero = FrameIO.LoadFrame(init.getFramesetNameAdjusted() + "0");
			processFrame(zero);
		} else {
			// get the frame path
			Frame zero = FrameIO.LoadFrame(_nameTo + "0");
			_framePath = zero.path;
			_firstNumber = _lastNumber + 1;
		}

		return super.initialise(init);
	}

	@Override
	protected void processFrame(Frame toProcess) {
		// load a fresh copy of the frame that bypasses the cache
		FrameIO.SuspendCache();

		Frame fresh = FrameIO.LoadFrame(toProcess.getFrameName());
		if (_nameMap.containsKey(fresh.getFrameName().toLowerCase())) {
			fresh
					.setFrameName(_nameMap.get(fresh.getFrameName()
							.toLowerCase()));
		} else {
			fresh.setFrameset(_nameTo);
			fresh.setFrameNumber(++_lastNumber);

			_nameMap.put(toProcess.getFrameName().toLowerCase(), fresh
					.getFrameName().toLowerCase());
		}

		boolean added = false;
		for (Item i : fresh.getItems())
			if (i.getLink() != null && !i.isAnnotation() && i.isLinkValid()) {
				String link = i.getLink();
				link = link.toLowerCase();
				// link = link.replace(_nameFrom, _nameTo);
				if (_nameMap.containsKey(link))
					link = _nameMap.get(link);
				else if (link.startsWith(_nameFrom)) {
					_nameMap.put(link, _nameTo + (++_lastNumber));
					link = _nameTo + _lastNumber;
				}
				i.setLink(link);
			} else if (!added && i.getLink() != null && i.isAnnotation()
					&& i.isLinkValid()) {
				// annotation links need to be parsed at the end
				if (i.getLink().toLowerCase().startsWith(_nameFrom)) {
					_toReparse.add(fresh.getFrameName());
					added = true;
				}

			}

		fresh.path = _framePath;
		FrameIO.SaveFrame(fresh);
		FrameIO.ResumeCache();
	}

	@Override
	protected void finalise(Frame frame) {
		// reparse all frames that have annotation links that may need updating
		for (String name : _toReparse) {
			Frame toParse = FrameIO.LoadFrame(name);
			boolean changed = false;
			for (Item i : toParse.getItems()) {
				if (i.getLink() != null && i.isAnnotation() && i.isLinkValid()) {
					String link = i.getLink();
					link = link.toLowerCase();
					// link = link.replace(_nameFrom, _nameTo);
					if (_nameMap.containsKey(link)) {
						link = _nameMap.get(link);
						i.setLink(link);
						changed = true;
					}
				}
			}

			if (changed)
				FrameIO.SaveFrame(toParse);
		}

		message("Tree successfully copied to " + _nameTo);
		FrameUtils.DisplayFrame(Frame.GetFramesetNameAdjusted(_nameTo) + _firstNumber);
	}

}
