package org.expeditee.gui;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.util.ArrayList;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import javax.swing.Timer;

import org.expeditee.actions.Actions;
import org.expeditee.actions.NavigationActions;
import org.expeditee.items.Constraint;
import org.expeditee.items.Dot;
import org.expeditee.items.Item;
import org.expeditee.items.ItemPermission;
import org.expeditee.items.ItemUtils;
import org.expeditee.items.Line;
import org.expeditee.items.Picture;
import org.expeditee.items.Text;
import org.expeditee.stats.SessionStats;

public class FrameMouseActions implements MouseListener, MouseMotionListener,
		MouseWheelListener {

	// TODO say where/how used
	private static final int MOUSE_WHEEL_THRESHOLD = 3;

	private static final int MINIMUM_RANGE_DEPRESS_TIME = 250;

	private static final int RECTANGLE_TO_POINT_THRESHOLD = 20;

	private static Date _lastMouseClickDate = new Date();

	public static final int LITTLE_MOUSE_PAUSE = 500;

	public static final int ZERO_MOUSE_PAUSE = 0;

	public static final int BIG_MOUSE_PAUSE = 750;

	public static final int CONTEXT_FREESPACE = 0;

	public static final int CONTEXT_AT_TEXT = 1;

	public static final int CONTEXT_AT_LINE = 2;

	public static final int CONTEXT_AT_DOT = 3;

	public static final int CONTEXT_AT_ENCLOSURE = 4;

	public static int _alpha = -1;

	/**
	 * The last known mouse X coordinate
	 */
	public static int MouseX;

	/**
	 * The last known mouse Y coordinate. Relative to the top of the
	 * application.
	 */
	public static int MouseY;

	// Offset coordinates from the mouse position when grabbing items
	private static int _offX;

	private static int _offY;

	// Keeps track of mouse button events when a delete occurs
	private static boolean _isDelete = false;

	// Keeps track of mouse button events when the user extracts attributes
	// occurs
	private static boolean _isAttribute = false;

	/**
	 * A flag to indicate that the last mouseUp event was part of a two button
	 * click sequence hence the next mouse up should be ignored*
	 */
	private static boolean _wasDouble = false;

	private static boolean _isNoOp = false;

	// keeps track of the last highlighted Item
	private static Item _lastHighlightedItem = null;

	// keeps track of the item being 'ranged out' if there is one.
	private static Text _lastRanged = null;

	// keeps track of the picture being cropped if there is one
	private static Picture _lastCropped = null;

	// true if lastItem only has highlighting removed when a new item is
	// highlighted
	private static boolean _lastHoldsHighlight = false;

	private static Item _tdfcItem = null;

	private static boolean _forceArrowCursor = true;

	// the current context of the cursor
	private static int _context = 0;

	public static void setForceArrow(boolean val) {
		_forceArrowCursor = val;
	}

	public static int getContext() {
		return _context;
	}

	private static MouseEvent _lastMouseClick = null;

	private static ItemPermission _lastClickedOn = null;

	private static List<ItemPermission> _lastClickedIn = null;

	private static boolean _pulseOn = false;

	private static int _pulseAmount = 2;

	private static Timer _MouseTimer = new Timer(LITTLE_MOUSE_PAUSE,
			new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					// check if we are in free space
					if (_lastClickedOn == null && Frame.FreeItems.size() == 0) {
						System.out.println("SuperBack!");
						leftButton(null);
						_MouseTimer.setDelay(ZERO_MOUSE_PAUSE);
					} else {
						_pulseOn = !_pulseOn;

						if (_MouseTimer.getDelay() == LITTLE_MOUSE_PAUSE)
							_MouseTimer.setDelay(LITTLE_MOUSE_PAUSE + 1);
						else
							_MouseTimer.setDelay(BIG_MOUSE_PAUSE);
						if (FrameUtils.getCurrentItem() == null) {
							for (Item i : Frame.FreeItems) {
								if (i instanceof Line) {
									Line line = (Line) i;
									if (_pulseAmount < 4
											&& line.getThickness() < 0)
										_pulseAmount = 4;

									if (_pulseOn)
										line.setThickness(line.getThickness()
												+ _pulseAmount);
									else
										line.setThickness(line.getThickness()
												- _pulseAmount);

									line.toggleArrow();
								}
							}
							FrameGraphics.Repaint();
						}
						_MouseTimer.stop();
					}
				}
			});

	public void mouseClicked(MouseEvent e) {
	}

	/**
	 * Each Item on the Frame is checked to determine if the mouse x,y
	 * coordinates are on the Item (or within the Shape surrounding it). If the
	 * coordinates are on the Item then the Item is checked for a link, if it
	 * has a link the link is followed, if not, nothing is done.
	 */
	public void mousePressed(MouseEvent e) {
		_lastMouseClickDate = new Date();
		_lastClickedOn = FrameUtils.getCurrentItem();
		// load any frame if necessary
		ItemPermission on = _lastClickedOn;

		_lastClickedIn = FrameUtils.getCurrentItems();

		SessionStats.MouseClicked(e.getButton());
		if (e.getButton() == MouseEvent.BUTTON1) {
			SessionStats.AddFrameEvent("Ld");
		} else if (e.getButton() == MouseEvent.BUTTON2) {
			SessionStats.AddFrameEvent("Md");
		} else if (e.getButton() == MouseEvent.BUTTON3) {
			SessionStats.AddFrameEvent("Rd");
		}

		_lastMouseClick = e;
		_MouseTimer.start();

		// pre-cache the frame if it is linked
		if (on != null && on.Item.getLink() != null && on.Item.isLinkValid()) {
			FrameIO.Precache(on.Item.getLink());
		}

		// check for delete command
		if (isDelete(e)) {
			_isDelete = true;
			// _lastRanged = null;
			_lastCropped = null;
			_wasDouble = false;
			// check for attributes command
		} else if (isGetAttributes(e)) {
			_isAttribute = true;
			_wasDouble = false;
		} else if (isTwoClickNoOp(e)) {
			_isAttribute = false;
			_wasDouble = false;
			_isDelete = false;
			_isNoOp = true;
		} else
			_isDelete = false;

		{
			// This must happen before the previous code
			if (_context == CONTEXT_FREESPACE && Frame.itemAttachedToCursor()) {
				_lastHighlightedItem = Frame.getItemAttachedToCursor();
			}
			// show depressed highlighting
			if (_lastHighlightedItem != null) {
				_lastHighlightedItem.showDepressedHighlight(true);
				FrameGraphics.Repaint();
			}
		}

		// if the user is ranging text
		if (on != null && on.Item instanceof Text && !_isDelete) {
			_lastRanged = (Text) on.Item;
			// set start-drag point
			_lastRanged.setSelectionStart(DisplayIO.getMouseX(), DisplayIO
					.getMouseY());
		}

		if (on != null && on.Item instanceof Picture
				&& e.getButton() == MouseEvent.BUTTON3 && !_isDelete) {
			_lastCropped = (Picture) on.Item;
			// set start crop point
			_lastCropped.setStartCrop(DisplayIO.getMouseX(), DisplayIO
					.getMouseY());
			_lastCropped.setShowCrop(true);
		}
	}

	// This is where all the processing happens
	public void mouseReleased(MouseEvent e) {
		updateCursor();

		Text lastRanged = _lastRanged;
		_lastRanged = null;
		// Dont do ranging if the user moves really quickly...
		// They are probably trying to pick something up in this case
		if (lastRanged != null) {
			long depressTime = (new Date()).getTime()
					- _lastMouseClickDate.getTime();
			// double changeInDistance =
			// e.getPoint().distance(_currentMouseClick.getPoint());
			// double speed = changeInDistance * 1000 / changeInTime;

			// System.out.println(depressTime);

			if (depressTime < MINIMUM_RANGE_DEPRESS_TIME
					|| lastRanged.getSelectedSize() <= 0) {// Text.MINIMUM_RANGED_CHARS)
				// {
				lastRanged.clearSelection();
				lastRanged = null;
			}
		}

		FrameUtils.ResponseTimer.restart();

		_MouseTimer.stop();
		_MouseTimer.setDelay(LITTLE_MOUSE_PAUSE);

		if (_pulseOn) {
			for (Item i : Frame.FreeItems) {
				if (i instanceof Line) {
					Line line = (Line) i;
					line.setThickness(line.getThickness() - _pulseAmount);
				}
			}
			_pulseOn = false;
		}

		// if the user was performing actions, stop now
		if (_pulseAmount < 0) {
			_pulseAmount = 2;
			return;
		}

		// reset pulse value
		_pulseAmount = 2;

		if (Actions.isAgentRunning()) {
			Actions.stopAgent();
			return;
		}

		// if the last action was a delete, then ignore the next mouseup
		if (_wasDouble) {
			_wasDouble = false;
			return;
		}

		/*
		 * if (_isNoOp) { if (e.getButton() != MouseEvent.NOBUTTON) { _isNoOp =
		 * false; _wasDouble = true; // lastRanged.clearSelection();
		 * FrameGraphics.Repaint(); return; } }
		 */

		// get whatever the user was pointing at
		ItemPermission clickedOn = _lastClickedOn;
		List<ItemPermission> clickedIn = _lastClickedIn;

		MouseX = e.getX();
		MouseY = e.getY();

		List<ItemPermission> releasedIn = FrameUtils.getCurrentItems();
		ItemPermission releasedOn = FrameUtils.getCurrentItem();

		if (_isNoOp
				&& ((releasedOn != null && clickedOn != null && releasedOn.Item != clickedOn.Item) || (releasedOn == null && releasedIn == null))) {
			if (_isDelete) {
				_isDelete = false;
				_wasDouble = true;
			}

			_isNoOp = false;

			if (_lastHighlightedItem != null)
				FrameGraphics.Highlight(_lastHighlightedItem, false, false);
			// TODO Michael thinks this code should never be reached and can be
			// removed
			if (Frame.itemAttachedToCursor()) {
				move(Frame.FreeItems);
			}

			FrameGraphics
					.DisplayMessage("Action cancelled, mouse moved more than "
							+ UserSettings.NoOpThreshold + " pixels.");
			FrameGraphics.Repaint();
			return;
		} else
			_isNoOp = false;

		// if this is a delete command
		if (_isDelete) {
			if (lastRanged != null) {
				lastRanged.cutSelectedText();
				lastRanged.clearSelection();
				FrameGraphics.Repaint();

			} else {
				delete(clickedOn);
			}
			_wasDouble = true;
			_isDelete = false;
			return;
		}

		// if this is an attribute extraction command
		if (_isAttribute) {
			extractAttributes(clickedOn);
			_wasDouble = true;
			_isAttribute = false;
			return;
		}

		// if the user is ranging-out text
		if (lastRanged != null) {

			Text ranged = lastRanged.copy();

			// if the user is cutting text from the item
			if (e.getButton() == MouseEvent.BUTTON2) {
				// if the entire text is selected then pickup the item
				if (lastRanged.getSelectedSize() == lastRanged.getLength()) {
					pickup(lastRanged);
					lastRanged.clearSelection();
					FrameGraphics.Repaint();
					return;
				}

				ranged.setText(lastRanged.cutSelectedText());
				// if the user is copying text from the item
			} else if (e.getButton() == MouseEvent.BUTTON3)
				ranged.setText(lastRanged.copySelectedText());

			ranged.setLink(null);
			ranged.setParent(null);
			ranged.setPosition(DisplayIO.getMouseX(), DisplayIO.getMouseY());
			pickup(ranged);

			lastRanged.clearSelection();
			FrameGraphics.Repaint();
			return;
		}

		// if the user is cropping an imate
		if (clickedOn != null && clickedOn.Item == _lastCropped
				&& _lastCropped.getCroppedSize() > 0) {
			Picture cropped = _lastCropped.copy();

			cropped.setParent(null);
			pickup(cropped);

		}

		// if the user has cropped an imate, either the above happend or this is
		// a no-op
		if (_lastCropped != null && _lastCropped.getCroppedSize() > 0) {
			_lastCropped.clearCropping();
			_lastCropped = null;
			FrameGraphics.Repaint();
			return;
		}

		// if the user is left-clicking
		if (e.getButton() == MouseEvent.BUTTON1) {
			SessionStats.AddFrameEvent("Lu");
			leftButton(clickedOn);
			return;
		}

		if (e.getButton() == MouseEvent.BUTTON2) {
			SessionStats.AddFrameEvent("Mu");
			middleButton(clickedOn, clickedIn);
			return;
		}

		if (e.getButton() == MouseEvent.BUTTON3) {
			SessionStats.AddFrameEvent("Ru");
			rightButton(clickedOn, clickedIn);
			return;
		}

		// error, we should have returned by now
		System.out.println("Error: mouseReleased should have returned by now. "
				+ e);
	}

	/**
	 * This method handles all left-click actions
	 */
	public static void leftButton(ItemPermission clicked) {
		// if the user is pointing at something then either follow the link or
		// do TDFC
		if (clicked != null) {
			// check item permissions
			if (clicked.Permission < ItemPermission.PERMISSION_FOLLOW_LINKS) {
				FrameGraphics
						.DisplayMessage("Insufficient Permissions for action on item");
				return;
			}

			Item clickedOn = clicked.Item;
			// actions take priority
			if (_lastMouseClick != null && !_lastMouseClick.isShiftDown()
					&& clickedOn.getAction() != null) {
				clickedOn.performActions();
				return;
			} else if (clickedOn.getLink() != null) {
				// Dont save the frame if we are moving to an old version of
				// this frame
				// because everytime we save with the old tag... the frame is
				// backed up
				if (!clickedOn.isOldTag())
					FrameIO.SaveFrame(DisplayIO.getCurrentFrame());

				NavigationActions.setLastNavigationItem(clickedOn);
				load(clickedOn.getAbsoluteLink());
				// DisplayIO.UpdateTitle();
				return;
				// no link is found, perform TDFC
			} else {
				// check for TDFC permission
				if (clicked.Permission < ItemPermission.PERMISSION_TDFC) {
					FrameGraphics
							.DisplayMessage("Insufficient permission to TDFC from that item");
					return;
				}

				if (clickedOn.isOldTag())
					return;

				// if the user is clicking on the frame name
				if (clickedOn.isFrameName()) {
					Frame next = FrameIO.LoadNext();
					FrameUtils.DisplayFrame(next, true);
					return;
				}

				tdfc(clickedOn);
				DisplayIO.UpdateTitle();
				return;
			}
			// if the user is not pointing at something, then this is a back
			// command
		} else {
			back();
			DisplayIO.UpdateTitle();
			return;
		}
	}

	private static boolean doMerging(ItemPermission clicked) {
		if (clicked == null)
			return false;

		// System.out.println(Frame.FreeItems.size());
		if (isRubberBandingRectangle()
				&& clicked.Item.getAllConnected().contains(
						Frame.getItemAttachedToCursor()))
			return true;

		if (Frame.FreeItems.size() > 2)
			return false;

		Item attachedToCursor = Frame.getItemAttachedToCursor();

		if (clicked.Item instanceof Text && !(attachedToCursor instanceof Text)) {
			return false;
		}

		return true;
	}

	/**
	 * This method handles all middle-click actions
	 */
	private static void middleButton(ItemPermission clicked,
			List<ItemPermission> clickedIn) {
		// if the cursor has Items attached
		if (Frame.itemAttachedToCursor()) {
			// if the user is pointing at something, merge the items (if
			// possible)
			if (doMerging(clicked)) {
				// check permissions
				if (clicked.Permission < ItemPermission.PERMISSION_FULL
						&& clicked.Item.getParent() != null
						&& clicked.Item.getParent().getName() != clicked.Item) {
					FrameGraphics.DisplayMessage("Insufficient Permission");
					return;
				}

				List<Item> left = merge(Frame.FreeItems, clicked.Item);
				anchor(left);

				Frame.FreeItems.clear();
				FrameGraphics.Repaint();
				updateCursor();
				return;
				// otherwise, anchor the items
			} else {
				// if a line is being rubber-banded, check for auto
				// straightening
				anchor(Frame.FreeItems);
				Frame.FreeItems.clear();
				updateCursor();
				_offX = _offY = 0;
				return;
			}
		} else {
			// otherwise if the user is pointing at something,pick it up
			if (clicked != null) {
				// check permissions
				if (clicked.Permission < ItemPermission.PERMISSION_FULL) {
					FrameGraphics
							.DisplayMessage("Insufficient Permission to pick up item");
					return;
				}

				pickup(clicked.Item);
				return;
				// otherwise create a line
			} else {
				if (clickedIn != null) {
					ArrayList<Item> toPickup = new ArrayList<Item>(clickedIn
							.size());
					for (ItemPermission ip : clickedIn)
						if (ip.Permission >= ItemPermission.PERMISSION_FULL)
							toPickup.add(ip.Item);

					pickup(toPickup);
					// otherwise the user is creating a line
				} else
					createLine();
				return;
			}
		}
	}

	private static Dot getFirstFreeDot() {
		for (Item i : Frame.FreeItems)
			if (i instanceof Dot)
				return (Dot) i;
		return null;
	}

	private static boolean isRubberBandingRectangle() {
		if (Frame.FreeItems.size() != 3)
			return false;
		Dot d = null;
		// only one dot will be present for rectangles
		for (Item i : Frame.FreeItems)
			if (i instanceof Dot && d == null)
				d = (Dot) i;
			else if (i instanceof Dot && d != null)
				d = null;

		// System.out.println(d.getAllConnected());

		// if this is indeed a rectangle anchor
		if (d != null && d.getAllConnected().size() > 5)
			return true;

		return false;
	}

	/**
	 * This method handles all right-click action
	 */
	private static void rightButton(ItemPermission clicked,
			List<ItemPermission> clickedIn) {
		// if the cursor has Items attached, then anchor a copy of them
		if (Frame.itemAttachedToCursor()) {
			// if the user is clicking on something, merge the items
			// unless it is a point onto somethin other than a point
			if (clicked != null
					&& (!(Frame.getItemAttachedToCursor() instanceof Dot && !(clicked.Item instanceof Dot)))) {
				// check permissions
				if (clicked.Permission < ItemPermission.PERMISSION_FULL
						&& clicked.Item.getParent().getName() != clicked.Item) {
					FrameGraphics
							.DisplayMessage("Insufficient Permission to merge items");
					return;
				}

				if (clicked.Item instanceof Text || clicked.Item instanceof Dot) {
					List<Item> copies = null;

					if (isRubberBandingRectangle()) {
						List<Item> remain = merge(Frame.FreeItems, clicked.Item);

						Dot d = getFirstFreeDot();
						// anchor the points
						anchor(remain);
						Frame.FreeItems.clear();
						updateCursor();
						// pick up a copy of all enclosed items
						List<ItemPermission> items = FrameUtils
								.getItemsEnclosedBy(
										DisplayIO.getCurrentFrame(), d
												.getEnclosedShape());
						if (items != null) {
							ArrayList<Item> toCopy = new ArrayList<Item>(items
									.size());
							for (ItemPermission ip : items)
								if (ip.Permission >= ItemPermission.PERMISSION_COPY)
									toCopy.add(ip.Item);

							copies = copy(toCopy);
							pickup(copies);
						}

					} else if (Frame.FreeItems.size() == 2) {
						copies = ItemUtils.UnreelLine(Frame.FreeItems);
						List<Item> leftOver = merge(Frame.FreeItems,
								clicked.Item);
						anchor(leftOver);
						if (copies == null)
							copies = copy(Frame.FreeItems);
						Frame.FreeItems.clear();
						for (Item i : copies)
							i.setOffset(0, 0);
						// need to move to prevent cursor dislocation
						move(copies);
						pickup(copies);

					} else if (Frame.FreeItems.size() == 1) {
						copies = copy(Frame.FreeItems);
						List<Item> remain = merge(copies, clicked.Item);

						// ignore items that could not be merged.
						anchor(remain);
					} else {
						// NoOp
					}
					updateCursor();
					FrameGraphics.Repaint();
				} else {
					List<Item> copies = ItemUtils.UnreelLine(Frame.FreeItems);
					if (copies == null)
						copies = copy(Frame.FreeItems);
					for (Item i : copies) {
						i.setOffset(0, 0);
					}
					anchor(Frame.FreeItems);
					Frame.FreeItems.clear();
					updateCursor();

					FrameGraphics.Repaint();

					clearParent(copies);
					pickup(copies);
				}
				return;

				// otherwise, anchor the items
			} else {

				// check if this is anchoring a rectangle
				if (isRubberBandingRectangle()) {
					Dot d = getFirstFreeDot();

					// anchor the points
					anchor(Frame.FreeItems);
					Frame.FreeItems.clear();
					updateCursor();
					// pick up a copy of all enclosed items
					List<ItemPermission> items = FrameUtils.getItemsEnclosedBy(
							DisplayIO.getCurrentFrame(), d.getEnclosedShape());
					if (items != null) {
						ArrayList<Item> toCopy = new ArrayList<Item>(items
								.size());
						for (ItemPermission ip : items)
							if (ip.Permission >= ItemPermission.PERMISSION_COPY)
								toCopy.add(ip.Item);
						toCopy.removeAll(d.getAllConnected());

						if (toCopy.size() > 0) {
							pickup(copy(toCopy));
							DisplayIO.getCurrentFrame().removeAllItems(d.getAllConnected());
						} else {
							pickup(copy(d.getAllConnected()));
						}
					}
					return;
				}
				List<Item> copies = null;
				if (Frame.FreeItems.size() == 2) {
					if (clicked != null) {
						List<Item> leftOver = merge(Frame.FreeItems,
								clicked.Item);
						anchor(leftOver);
					}
					// This is executed when the user is putting down a line
					// endpoint and unreeling. ie. Normal unreeling
					copies = ItemUtils.UnreelLine(Frame.FreeItems);

					if (copies == null)
						copies = copy(Frame.FreeItems);
					anchor(Frame.FreeItems);
					for (Item i : copies)
						i.setOffset(0, 0);
					// need to move to prevent cursor dislocation
					move(copies);
					pickup(copies);
					FrameGraphics.Repaint();
					return;
				}

				if (copies == null)
					copies = copy(Frame.FreeItems);
				// MIKE: what does the below 2 lines do?
				for (Item i : copies)
					i.setOffset(0, 0);
				// The below code has a little problem withflicker when stamp
				// and dragging
				move(Frame.FreeItems);
				anchor(copies);
				updateCursor();
				FrameGraphics.Repaint();
				return;
			}

		} else {
			// if the user is pointing at something, this is a copy
			if (clicked != null) {
				// check permissions
				if (clicked.Permission < ItemPermission.PERMISSION_COPY) {
					FrameGraphics
							.DisplayMessage("Insufficient permission to copy item");
					return;
				}

				// This happens when the user right clicks on the endpoint of an
				// existing line
				// check for unreeling
				List<Item> copies = ItemUtils.UnreelLine(clicked.Item);

				if (copies == null) {
					// check if this is picking up a rectangle
					if (clicked.Item instanceof Dot
							&& clicked.Item.getConnected().size() == 3) {
						Dot d = null;
						// only one dot will be present for rectangles
						for (Item i : clicked.Item.getConnected())
							if (i instanceof Dot && d == null)
								d = (Dot) i;
							else if (i instanceof Dot && d != null)
								d = null;

						// if this is indeed a rectangle anchor
						if (d != null) {
							// pick up a copy of all enclosed items
							List<ItemPermission> items = FrameUtils
									.getItemsEnclosedBy(DisplayIO
											.getCurrentFrame(), d
											.getEnclosedShape());
							if (items != null) {
								ArrayList<Item> toCopy = new ArrayList<Item>(
										items.size());
								for (ItemPermission ip : items)
									if (ip.Permission >= ItemPermission.PERMISSION_COPY)
										toCopy.add(ip.Item);

								List<Item> toPickup = copy(toCopy);
								clearParent(toPickup);
								pickup(toPickup);
							}
							return;
						}
					}
					copies = copy(clicked.Item);
				}

				pickup(copies);
				return;

			} else {
				// if the user is pointing inside a closed shape, copy the items
				// inside
				if (clickedIn != null) {
					List<Item> toCopy = new ArrayList<Item>(clickedIn.size());
					for (ItemPermission ip : clickedIn)
						if (ip.Permission >= ItemPermission.PERMISSION_COPY)
							toCopy.add(ip.Item);

					List<Item> copies = copy(toCopy);
					clearParent(copies);
					pickup(copies);
					// otherwise, create a rectangle
				} else {

					int x = DisplayIO.getMouseX();
					int y = DisplayIO.getMouseY();

					// create dots
					Dot d1 = new Dot(x, y, DisplayIO.getCurrentFrame()
							.getNextItemID());
					Dot d2 = new Dot(x, y, DisplayIO.getCurrentFrame()
							.getNextItemID());
					Dot d3 = new Dot(x, y, DisplayIO.getCurrentFrame()
							.getNextItemID());
					Dot d4 = new Dot(x, y, DisplayIO.getCurrentFrame()
							.getNextItemID());

					// create lines
					Line l1 = new Line(d1, d2, DisplayIO.getCurrentFrame()
							.getNextItemID());
					Line l2 = new Line(d2, d3, DisplayIO.getCurrentFrame()
							.getNextItemID());
					Line l3 = new Line(d3, d4, DisplayIO.getCurrentFrame()
							.getNextItemID());
					Line l4 = new Line(d4, d1, DisplayIO.getCurrentFrame()
							.getNextItemID());

					new Constraint(d1, d2, DisplayIO.getCurrentFrame()
							.getNextItemID(), Constraint.HORIZONTAL);
					new Constraint(d3, d4, DisplayIO.getCurrentFrame()
							.getNextItemID(), Constraint.HORIZONTAL);
					new Constraint(d2, d3, DisplayIO.getCurrentFrame()
							.getNextItemID(), Constraint.VERTICAL);
					new Constraint(d4, d1, DisplayIO.getCurrentFrame()
							.getNextItemID(), Constraint.VERTICAL);

					anchor(d1);
					anchor(d2);
					anchor(d3);
					anchor(d4);

					anchor(l1);
					anchor(l2);
					anchor(l3);
					anchor(l4);

					pickup(d4);

					FrameGraphics.Repaint();
					return;
				}
			}
		}
	}

	private static void clearParent(List<Item> items) {
		for (Item i : items)
			i.setParent(null);
	}

	public void mouseEntered(MouseEvent e) {
	}

	public void mouseExited(MouseEvent e) {
	}

	public void mouseDragged(MouseEvent e) {
		// System.out.println("MouseDragged");

		// check if user is dragging across a text item
		if (_lastRanged != null) {
			// System.out.println(MouseY - e.getY());

			MouseX = e.getX();
			MouseY = e.getY();

			int distance = _lastRanged.getY() - DisplayIO.getMouseY();
			if (distance <= 0)
				distance = DisplayIO.getMouseY() - _lastRanged.getY()
						- _lastRanged.getBoundsHeight();

			if (distance > UserSettings.NoOpThreshold) {
				_lastRanged.clearSelectionEnd();
				_isNoOp = true;
			} else {
				// update the ranged section
				_lastRanged.setSelectionEnd(DisplayIO.getMouseX(), DisplayIO
						.getMouseY());
				_isNoOp = false;
			}

			DisplayIO.setTextCursor(_lastRanged.getSize());
			FrameGraphics.Repaint();
			return;
		}

		// if the user is dragging across a picture
		if (_lastCropped != null) {
			MouseX = e.getX();
			MouseY = e.getY();

			// update the ranged section
			_lastCropped.setEndCrop(DisplayIO.getMouseX(), DisplayIO
					.getMouseY());
			DisplayIO.setCursor(Item.CROP_CURSOR);
			FrameGraphics.Repaint();
			return;
		}

		/*
		 * // Euclidean distance int distance = (int)
		 * Math.sqrt(Math.pow(Math.abs(MouseX - e.getX()), 2) +
		 * Math.pow(Math.abs(MouseY - e.getY()), 2));
		 * 
		 * if (distance > UserSettings.NoOpThreshold)
		 */
		// Use the below calculation for better speed. If it causes problems
		// switch back to the Euclidean distance calculation
		if (Math.abs(MouseX - e.getX()) > UserSettings.NoOpThreshold
				|| Math.abs(MouseY - e.getY()) > UserSettings.NoOpThreshold)
			_isNoOp = true;
	}

	MouseEvent _lastMouseEvent = null;

	/**
	 * Updates the stored mouse position and highlights any items as necessary.
	 */
	public void mouseMoved(MouseEvent e) {
		// System.out.println("MouseMoved");

		MouseX = e.getX();
		MouseY = e.getY();

		if (_lastMouseEvent == null)
			_lastMouseEvent = e;

		/*
		 * int distance = (int) Math.sqrt(Math.pow(Math
		 * .abs(_lastMouseEvent.getX() - e.getX()), 2) +
		 * Math.pow(Math.abs(_lastMouseEvent.getY() - e.getY()), 2));
		 */

		// Just do the sum of the distance moved in x and y direction for
		// speed!!
		int distance = Math.abs(_lastMouseEvent.getX() - e.getX())
				+ Math.abs(_lastMouseEvent.getY() - e.getY());

		boolean checkHighlight = true;

		// Mike: This code checks if the mouse is moving above a
		// threshold speed. If so it doesnt execute the code that checks if
		// items underneath the cursor need to be highlighted.
		if (Frame.itemAttachedToCursor()) {
			long time = e.getWhen() - _lastMouseEvent.getWhen();
			// System.out.println(time + " D:" + distance);
			if (time > 0 && distance * 3 > time) {
				checkHighlight = false;
				// System.out.println("Dont check highlight!");
				if (_lastHighlightedItem != null && !_lastHoldsHighlight) {
					FrameGraphics.Highlight(_lastHighlightedItem, false, true);
				}
			}
		}

		_lastMouseEvent = e;

		if (checkHighlight) {
			// ByMike: Get the item the mouse is hovering over
			ItemPermission click = FrameUtils.getCurrentItem();
			Item on = null;

			if (click != null) {
				on = click.Item;
				// set the context
				if (on instanceof Line)
					_context = CONTEXT_AT_LINE;
				else if (on instanceof Dot)
					_context = CONTEXT_AT_DOT;
				else if (on instanceof Text) {
					// Checks that we are actually pointing on a character
					// not just space in the text box's bounding box
					if (((Text) on).contains(DisplayIO.getMouseX(), DisplayIO
							.getMouseY())) {
						_context = CONTEXT_AT_TEXT;
					} else {
						// if we are pointing in space in the text box... it is
						// the same as pointing in freespace
						on = null;
						_context = CONTEXT_FREESPACE;
					}
				}

				_alpha = 60;
			} else {
				_context = CONTEXT_FREESPACE;
				_alpha = -1;
			}

			// if the user is pointing at an item, highlight it
			if (on != null && !Frame.FreeItems.contains(on)) {
				// if the user can spot-weld, show the virtual spot
				if (Frame.FreeItems.size() == 2 && on instanceof Line) {
					Line line = (Line) on;
					if (Frame.FreeItems.get(0) instanceof Dot)
						line.showVirtualSpot((Dot) Frame.FreeItems.get(0),
								DisplayIO.getMouseX(), DisplayIO.getMouseY());
					else if (Frame.FreeItems.get(1) instanceof Dot)
						line.showVirtualSpot((Dot) Frame.FreeItems.get(1),
								DisplayIO.getMouseX(), DisplayIO.getMouseY());
					else
						FrameGraphics.Highlight(on, true, true);
				} else
					FrameGraphics.Highlight(on, true, true);

				// if the last item highlighted is still highlighted, clear it
				if (_lastHoldsHighlight) {
					_lastHoldsHighlight = false;
					for (Item i : DisplayIO.getCurrentFrame().getItems())
						if (i.isHighlighted() && i != on)
							FrameGraphics.Highlight(i, false, false);
				}

				// if the user is not pointing at an item, check for enclosure
				// highlighting
			} else if (on == null) {
				List<Dot> enclosure = FrameUtils.getEnclosingDots();
				if (enclosure != null && enclosure.size() > 1) {
					if (enclosure.get(0).getLines().size() > 1) {
						on = enclosure.get(0).getLines().get(0);
						// FrameGraphics.Highlight(enclosure.get(0).getLines().get(0),
						// true, true);
						FrameGraphics.Highlight(on, true, true);
					}
					// for(Dot d : enclosure){
					// FrameGraphics.Highlight(d, true, true);
					// break;
					// }
				} else if (_lastHighlightedItem != null) {
					_lastHoldsHighlight = false;
				}
			}

			// disable cursor changes when the cursor has items attached
			if (Frame.itemAttachedToCursor()
					&& DisplayIO.getCursor() != Item.TEXT_CURSOR)
				_forceArrowCursor = false;

			if (_lastHighlightedItem != null && _lastHighlightedItem != on
					&& !_lastHoldsHighlight) {
				FrameGraphics.Highlight(_lastHighlightedItem, false, true);
			}

			_lastHighlightedItem = on;

		}

		if (Frame.itemAttachedToCursor()) {
			move(Frame.FreeItems);
		}

		if (_forceArrowCursor)
			updateCursor();

		// if (_lastItem == null)
		// DisplayIO.UpdateTitle();

		_forceArrowCursor = true;
	}

	/**
	 * Updates the current mouse cursor to whatever it should be. i.e. Hidden
	 * when rubber-banding lines, otherwise default (arrow)
	 */
	private static void updateCursor() {
		// if rubber-banding, there will be exactly 2 items on the cursor
		if (Frame.FreeItems.size() != 2)
			DisplayIO.setCursor(Item.DEFAULT_CURSOR);
		else {
			Dot d = null;
			// check that it is indeed a line being rubber-banded
			if (Frame.FreeItems.get(0) instanceof Line
					&& Frame.FreeItems.get(1) instanceof Dot)
				d = (Dot) Frame.FreeItems.get(1);
			else if (Frame.FreeItems.get(1) instanceof Line
					&& Frame.FreeItems.get(0) instanceof Dot)
				d = (Dot) Frame.FreeItems.get(0);

			if (d != null) {
				DisplayIO.setCursor(Item.HIDDEN_CURSOR);
			} else
				DisplayIO.setCursor(Item.DEFAULT_CURSOR);
		}

	}

	public static void setHighlightHold(boolean hold) {
		_lastHoldsHighlight = hold;
	}

	public static void resetOffset() {
		if (Frame.itemAttachedToCursor()) {
			_offX = DisplayIO.getMouseX() - Frame.FreeItems.get(0).getX()
					+ Frame.FreeItems.get(0).getOffset().x;
			_offY = MouseY - Frame.FreeItems.get(0).getY()
					+ Frame.FreeItems.get(0).getOffset().y;
		}
	}

	private static void move(List<Item> toMove) {
		int deltax = toMove.get(0).getX() - (DisplayIO.getMouseX() - _offX);
		int deltay = toMove.get(0).getY() - (MouseY - _offY);

		for (Item move : toMove) {
			move.setPosition(move.getX() - deltax, move.getY() - deltay);

			if (move instanceof Text)
				((Text) move).setAlpha(_alpha);
		}

		FrameGraphics.Repaint();
	}

	private static void load(String toLoad) {
		checkTDFCItemWaiting();
		FrameUtils.DisplayFrame(toLoad);
	}

	private static void checkTDFCItemWaiting() {
		// if there is a TDFC Item waiting
		if (_tdfcItem != null) {
			Frame currentFrame = DisplayIO.getCurrentFrame();
			boolean change = currentFrame.hasChanged();
			boolean saved = currentFrame.isSaved();
			// Save the parent of the item if it has not been saved
			// TODO change it so it doesnt use load frame here... if
			// possible
			if (!change && !saved) {
				_tdfcItem.setLink(null);
				_tdfcItem.getParent().setChanged(true);
				FrameIO.SaveFrame(_tdfcItem.getParent());
				FrameGraphics.Repaint();
			} else {
				SessionStats.CreatedFrame();
				// DisplayIO.UpdateTitle();
			}

			_tdfcItem = null;
		}
	}

	private static void back() {
		checkTDFCItemWaiting();
		DisplayIO.Back();

		// repaint things if necessary
		if (Frame.itemAttachedToCursor())
			move(Frame.FreeItems);
	}

	/**
	 * Returns true if the mouse moved during TDFC. This will happen if there is
	 * a start annotation item on the frame.
	 * 
	 * @param linker
	 * @return
	 */
	public static boolean tdfc(Item linker) {
		if (linker instanceof Line)
			return false;

		// if this is a non-usable item
		if (linker.getID() < 0)
			return false;

		linker.getParent().setChanged(true);
		checkTDFCItemWaiting();

		Frame next = FrameIO.CreateNewFrame(linker);

		linker.setLink("" + next.getFrameNumber());
		_tdfcItem = linker;

		FrameUtils.DisplayFrame(next, true);

		boolean mouseMoved = next.moveMouseToDefaultLocation();
		// this needs to be done if the user doesnt move the mouse before doing
		// tdfc while the cursor is set to the text cursor
		DisplayIO.setCursor(Item.DEFAULT_CURSOR);
		// This needs to be done in case there was a @start on the frame which
		// triggers changed to be set to true when it should stay as false
		next.setChanged(false);

		return mouseMoved;
	}

	/**
	 * Creates a new Text item and fills it with particular attributes extracted
	 * from the given Item. Note: Users always have permission to extract
	 * attributes, so it is not checked.
	 * 
	 * @param toExtract
	 *            ItemPermission containing the Item to extract the attributes
	 *            from.
	 */
	private static void extractAttributes(ItemPermission toExtract) {
		if (toExtract == null || toExtract.Item == null)
			return;

		if (Frame.itemAttachedToCursor())
			return;

		Text attribs;
		Item item = toExtract.Item;
		// Extract the frames attributes when the user clicks on the frame name
		if (item.isFrameName())
			attribs = AttributeUtils.ExtractAttributes(item.getParent());
		else
			attribs = AttributeUtils.ExtractAttributes(item);

		if (attribs == null)
			FrameGraphics
					.DisplayMessage("All attributes of that item are default values.");
		else {
			// Give the attribute text item the color of the item for which
			// attributes are being extracted.
			attribs.setColor(item.getColor());
			pickup(attribs);
		}
	}

	private static void delete(ItemPermission toDelete) {
		FrameUtils.LastEdited = null;
		_offX = _offY = 0;

		// check if the user is pointing at the frame's framename
		if (toDelete != null
				&& toDelete.Item == DisplayIO.getCurrentFrame().getName()) {
			DisplayIO.getCurrentFrame().clear();
			FrameGraphics.Repaint();
			return;
		}

		// if the user is deleting items attached to the cursor
		if (Frame.itemAttachedToCursor()) {
			// if this is an item-swap
			if (Frame.FreeItems.size() == 1
					&& Frame.FreeItems.get(0) instanceof Text
					&& toDelete != null && toDelete.Item instanceof Text) {

				// check permissions
				if (toDelete.Permission < ItemPermission.PERMISSION_FULL) {
					FrameGraphics
							.DisplayMessage("Insufficient permission to swap Item text");
					return;
				}

				Text anchored = (Text) toDelete.Item;
				Text free = (Text) Frame.FreeItems.get(0);

				// List<String> temp = anchored.getText();
				anchored.setTextList(free.getText());

				// free.setTextList(temp);
				Frame.FreeItems.clear();

				anchored.getParent().setChanged(true);

				// update the offset since the text has changed
				_offX = DisplayIO.getMouseX() - anchored.getX()
						+ anchored.getOffset().x;
				_offY = MouseY - anchored.getY() + anchored.getOffset().y;
			} else {
				deleteItems(Frame.FreeItems);
			}
			// reset the mouse cursor
			updateCursor();
			FrameGraphics.Repaint();
			return;
			// the user is not pointing at an item
		} else if (toDelete == null) {

			// if the user is pointing inside a closed shape, delete all
			// items inside it
			List<ItemPermission> items = FrameUtils.getCurrentItems();

			if (items != null) {
				ArrayList<Item> toRemove = new ArrayList<Item>(items.size());
				for (ItemPermission ip : items)
					if (ip.Permission >= ItemPermission.PERMISSION_FULL) {
						toRemove.add(ip.Item);
					}

				deleteItems(toRemove);

				// reset the mouse cursor
				updateCursor();
				FrameGraphics.Repaint();

				// otherwise this is an undo command
			} else {
				DisplayIO.getCurrentFrame().undo();
			}
			return;
			// this is a delete command
		} else {
			// check permissions
			if (toDelete.Permission < ItemPermission.PERMISSION_FULL) {
				FrameGraphics
						.DisplayMessage("Insufficient permission to delete item");
				return;
			}

			Frame current = toDelete.Item.getParent();

			current.setChanged(true);

			List<Item> toUndo = new LinkedList<Item>();

			if (toDelete.Item instanceof Dot) {
				toUndo.addAll(deleteDot((Dot) toDelete.Item));
			} else
				toUndo.addAll(copy(toDelete.Item.getConnected()));

			current.addAllToUndo(toUndo);
			current.removeAllItems(toDelete.Item.getConnected());
			// reset the mouse cursor
			updateCursor();
			FrameGraphics.Repaint();

			ItemUtils.EnclosedCheck(current.getItems());
			FrameGraphics.Repaint();
		}
		// check
		DisplayIO.setCursor(Item.DEFAULT_CURSOR);
	}

	private static void deleteItems(List<Item> itemList) {
		Frame current = DisplayIO.getCurrentFrame();
		List<Item> toUndo = new LinkedList<Item>();
		List<Item> seen = new LinkedList<Item>();

		// disconnect any connected items
		for (Item i : itemList) {
			if (i.getLines().size() > 0) {

				// check for rectangle deletion
				if (itemList.size() == 3) {
					boolean rectangle = false;
					// check that this is indeed a rectangle
					for (Item check : itemList)
						if (check != i && !i.getLines().contains(check)) {
							rectangle = false;
							break;
						} else
							rectangle = true;

					// if this is a rectangle, remove the entire
					// rectangle
					if (rectangle) {
						List<Item> copy = copy(i.getAllConnected());
						current.addAllToUndo(copy);
						current.removeAllItems(i.getAllConnected());
						itemList.clear();
						// reset the mouse cursor
						updateCursor();
						FrameGraphics.Repaint();
						return;
					}
				}

				List<Item> copy = deleteDot((Dot) i);
				// add the copied items to the undo stack
				for (Item cop : copy)
					if (!toUndo.contains(cop))
						toUndo.add(cop);
			} else {
				if (!seen.contains(i))
					toUndo.addAll(copy(i));
			}

			seen.add(i);
		}

		// check for rectangle deletion
		if (itemList.size() == 3) {
			boolean rectangle = false;
			Dot dot = null;
			// check that this is indeed a rectangle
			for (Item check : itemList) {
				if (dot == null && check instanceof Dot)
					dot = (Dot) check;
				else if (check != dot && !dot.getLines().contains(check)) {
					rectangle = false;
					break;
				} else
					rectangle = true;
			}

			// if this is a rectangle, remove the entire rectangle
			if (rectangle) {
				// List<Item> copy = copy(dot.getAllConnected());
				current.addAllToUndo(dot.getAllConnected());
				current.removeAllItems(dot.getAllConnected());
				itemList.clear();
				// reset the mouse cursor
				updateCursor();
				FrameGraphics.Repaint();
				return;
			}
		}

		current.removeAllItems(itemList);
		current.addAllToUndo(itemList);
		itemList.clear();
	}

	private static List<Item> deleteDot(Dot dot) {
		// create a backup copy of the dot
		List<Item> copy = copy(dot.getConnected());
		// remove lines that are connected to anchored dots
		// note: the line is kept so that it can be properly restored
		for (Item ic : copy) {
			if (ic instanceof Line) {
				Line line = (Line) ic;
				if (!copy.contains(line.getStartItem()))
					line.getStartItem().removeLine(line);
				if (!copy.contains(line.getEndItem()))
					line.getEndItem().removeLine(line);
			}
		}

		// remove all lines being deleted
		for (Item ic : dot.getConnected()) {
			if (ic instanceof Line && ((Line) ic).getOppositeEnd(dot) != null) {
				Line line = (Line) ic;

				Item d = line.getOppositeEnd(dot);
				d.removeLine(line);

				// if the dot was only part of one line, it can be
				// removed
				if (d.getLines().size() == 0) {
					if (d.getParent() != null)
						d.getParent().removeItem(d);
					if (!copy.contains(d))
						copy.add(d);
				}

				if (dot.getParent() != null)
					dot.getParent().removeItem(ic);
			}
		}
		return copy;
	}

	private static void removeAllLinesExcept(Item from, Item but) {
		List<Line> lines = new LinkedList<Line>();
		lines.addAll(from.getLines());
		for (Line line : lines)
			if (line.getOppositeEnd(from) != but)
				from.removeLine(line);

		List<Constraint> consts = new LinkedList<Constraint>();
		consts.addAll(from.getConstraints());
		for (Constraint c : consts)
			if (c.getOppositeEnd(from) != but)
				from.removeConstraint(c);
	}

	public static List<Item> merge(List<Item> merger, Item mergee) {
		if (mergee.getParent().getName() == mergee) {
			return mergee.getParent().merge(merger);
		}

		// check for rectangle merging
		if (merger.size() == 3 && mergee.getLines().size() == 2) {
			boolean rectangle = false;
			Dot dot = null;
			// check that this is indeed a rectangle
			for (Item check : merger) {
				if (dot == null && check instanceof Dot)
					dot = (Dot) check;
				else if (check != dot && !dot.getLines().contains(check)) {
					rectangle = false;
					break;
				} else
					rectangle = true;
			}

			// if this is a rectangle
			if (rectangle) {
				List<Item> connected = dot.getAllConnected();
				DisplayIO.getCurrentFrame().removeAllItems(connected);

				// find the point opposite the point that was clicked
				// (mergee)...
				Dot opposite = null;
				List<Line> lines = dot.getLines();
				for (Line l : lines) {
					if (l.getOppositeEnd(dot) != mergee) {
						opposite = (Dot) l.getOppositeEnd(dot);
						break;
					}
				}

				// check if the rectangle is small enough that it should be
				// collapsed to a single point
				int x1 = Math.abs(opposite.getX() - mergee.getX());
				int x2 = Math.abs(opposite.getY() - mergee.getY());
				int distance = (int) Math.sqrt(Math.pow(x1, 2)
						+ Math.pow(x2, 2));

				if (distance < RECTANGLE_TO_POINT_THRESHOLD) {
					mergee.removeAllConstraints();
					mergee.removeAllLines();
					return mergee.getAllConnected();
				} else {

					// otherwise, replace the rectangle with a constrained
					// line
					Item other = null;
					// there should be only one line left
					for (Line line : mergee.getLines())
						if (!merger.contains(line.getOppositeEnd(mergee)))
							other = line.getOppositeEnd(mergee);

					removeAllLinesExcept(mergee, other);
					removeAllLinesExcept(other, mergee);

					return mergee.getAllConnected();
				}
			}
		}

		List<Item> remain = new ArrayList<Item>();
		Item res = null;

		for (Item i : merger) {
			// check for link merging
			if (i instanceof Text
					&& FrameIO.isValidFrameName((((Text) i).getFirstLine()))
					&& FrameIO.DoesFrameExist((((Text) i).getFirstLine()))) {
				// check that we can actually access the frame this link
				// corresponds to

				mergee.setLink(((Text) i).getFirstLine());
			} else {
				// check for attribute merging
				if (i instanceof Text) {
					Text txt = (Text) i;

					// if this is not an attribute merge
					if (!AttributeUtils.SetAttribute(mergee, txt)) {
						// set mouse position for text merges
						if (mergee instanceof Text)
							// 18Feb Rob doesnt want cursor to jump when
							// inserting text
							((Text) mergee).insertText(txt.getTextNoList(),
									DisplayIO.getMouseX(), DisplayIO
											.getMouseY());

						else if (mergee instanceof Dot) {
							mergee.getParent().removeItem(mergee);
							break;
						}

						// TODO tidy this up... the text override of this method
						// doesnt use the parametres MouseY and MouseX!!
						res = mergee.merge(i, DisplayIO.getMouseX(), MouseY
								- DisplayIO.getOffset());
						if (res != null) {
							remain.add(res);
						}
					}
				} else {
					res = mergee.merge(i, DisplayIO.getMouseY(), MouseY
							- DisplayIO.getOffset());
					if (res != null)
						remain.addAll(res.getConnected());

				}
			}

		}

		updateCursor();
		mergee.getParent().setChanged(true);
		ItemUtils.EnclosedCheck(mergee.getParent().getItems());
		FrameUtils.Parse(mergee.getParent());

		return remain;
	}

	/**
	 * Picks up an item on a frame.
	 * 
	 * @param toGrab
	 *            item to be picked up
	 * @param removeItem
	 *            true if the item should be removed from the frame
	 */
	public static void pickup(Item toGrab) {
		pickup(toGrab.getConnected());
	}

	public static void pickup(List<Item> toGrab) {
		for (Item i : toGrab) {
			// Check if it has a relative link if so make it absolute
			i.setAbsoluteLink();

			// parent may be null
			if (i.getParent() != null)
				i.getParent().removeItem(i);

			Frame.FreeItems.add(i);
			i.setParent(null);
			if (i instanceof Dot)
				((Dot) i).setFloating(true);
		}

		updateCursor();

		// if there are multiple items in the list, determine which to use for
		// offset calculations
		if (toGrab.size() > 1) {
			for (Item i : toGrab) {
				if (!(i instanceof Line)) {
					_offX = DisplayIO.getMouseX() - i.getX() + i.getOffset().x;
					_offY = MouseY - i.getY() + i.getOffset().y;

					// make the offset item the first item in the list (so
					// moving code knows which item to use)
					Frame.FreeItems.set(Frame.FreeItems.indexOf(i),
							Frame.FreeItems.get(0));
					Frame.FreeItems.set(0, i);
					break;
				}
			}

			move(Frame.FreeItems);
			// otherwise, just use the first item
		} else if (toGrab.size() == 1) {
			_offX = DisplayIO.getMouseX() - toGrab.get(0).getX()
					+ toGrab.get(0).getOffset().x;
			_offY = MouseY - toGrab.get(0).getY() + toGrab.get(0).getOffset().y;
		}

		ItemUtils.EnclosedCheck(toGrab);
	}

	private static void createLine() {
		// create the two endpoints
		Dot end = new Dot(DisplayIO.getRealMouseX(), DisplayIO.getMouseY(),
				DisplayIO.getCurrentFrame().getNextItemID());
		Dot start = new Dot(DisplayIO.getRealMouseX(), DisplayIO.getMouseY(),
				DisplayIO.getCurrentFrame().getNextItemID());
		// create the Line
		Line line = new Line(start, end, DisplayIO.getCurrentFrame()
				.getNextItemID());
		line.autoArrowheadLength();

		// anchor the start
		anchor(start);

		// attach the line to the cursor
		pickup(end);
	}

	private static List<Item> copy(Item toCopy) {
		return copy(toCopy.getConnected());
	}

	/**
	 * Returns a list of copies of the list passed in
	 * 
	 * @param toCopy
	 *            The list of items to copy
	 * @return A List of copied Items
	 */
	private static List<Item> copy(List<Item> toCopy) {
		return ItemUtils.CopyItems(toCopy);
	}

	public static void anchor(Item toAnchor) {
		Frame current = DisplayIO.getCurrentFrame();
		// toAnchor.anchor(current);

		// update the items ID to prevent conflicts with the current frame
		if (toAnchor.getID() < 0 || current.getItems().contains(toAnchor))
			;
		toAnchor.setID(current.getNextItemID());

		toAnchor.setParent(current);
		toAnchor.setOffset(0, 0);
		current.addItem(toAnchor);
		current.setResort(true);

		toAnchor.setRelativeLink();

		if (toAnchor instanceof Dot) {
			for (Line line : ((Dot) toAnchor).getLines()) {
				if (line.getID() < 0 && !current.getItems().contains(line)) {
					line.setID(current.getNextItemID());
					line.showHighlight(true);
					// Mike: Why was this line here?
					// anchor(line);
				}
			}
			((Dot) toAnchor).setFloating(false);
		} else if (toAnchor instanceof Text) {
			// ensure all text items have their selection cleared
			((Text) toAnchor).clearSelection();
			((Text) toAnchor).setAlpha(0);
		} else if (toAnchor instanceof Line) {
			((Line) toAnchor).fixArrowheadLength();
		}

		FrameGraphics.Repaint();
		ItemUtils.EnclosedCheck(current.getItems());
	}

	public static void anchor(List<Item> toAnchor) {
		for (Item i : toAnchor)
			anchor(i);
		toAnchor.clear();
	}

	/*
	 * private boolean mouseMovedRecently() { Date now = new Date();
	 * 
	 * return now.getTime() - _lastMouseMovement.getTime() < 150; }
	 */

	public void mouseWheelMoved(MouseWheelEvent arg0) {
		// System.out.println(arg0.getClickCount() + " " +
		// arg0.getWheelRotation());
		int clicks = arg0.getClickCount();

		// if a line is being rubber-banded, check for auto straightening
		if (Frame.FreeItems.size() == 2) {
			if ((Frame.FreeItems.get(0) instanceof Dot && Frame.FreeItems
					.get(1) instanceof Line)
					|| (Frame.FreeItems.get(1) instanceof Dot && Frame.FreeItems
							.get(0) instanceof Line)) {

				Line line;
				if (Frame.FreeItems.get(0) instanceof Line)
					line = (Line) Frame.FreeItems.get(0);
				else
					line = (Line) Frame.FreeItems.get(1);

				// User must do multiple clicks to toggle the line
				if (clicks == MOUSE_WHEEL_THRESHOLD) {
					if (arg0.isShiftDown())
						line.toggleArrowHeadRatio(arg0.getWheelRotation());
					else
						line.toggleArrowHeadLength(arg0.getWheelRotation());
					FrameGraphics.Repaint();
				}
			}
		} else if (arg0.getWheelRotation() != 0 && arg0.isShiftDown()) {

			int rotationType = FrameKeyboardActions.SIZE_UP;
			if (arg0.getWheelRotation() > 0) {
				rotationType = FrameKeyboardActions.SIZE_DOWN;
			}

			for (int i = 0; i < clicks; i++)
				FrameKeyboardActions.functionKey(rotationType);
		} else if (clicks == MOUSE_WHEEL_THRESHOLD) {
			ItemPermission ip = FrameUtils.getCurrentItem();

			// if the user is not pointing to any item
			if (ip == null) {
				FrameGraphics
						.DisplayMessage("There are no Items selected on the Frame");
				return;
			} else {
				// check permissions
				if (ip.Permission < ItemPermission.PERMISSION_FULL) {
					FrameGraphics
							.DisplayMessage("Insufficient permission to change the size of that item");
					return;
				}
			}

			if (ip.Item instanceof Line) {
				Line line = (Line) ip.Item;
				line.toggleDashed(arg0.getWheelRotation());
				FrameGraphics.Repaint();
				return;
			}
		}
	}

	/**
	 * 
	 * @return the integer value for the last mouse button clicked.
	 */
	public static int getLastMouseButton() {
		if (_lastMouseClick == null)
			return MouseEvent.NOBUTTON;

		return _lastMouseClick.getButton();
	}

	public static boolean isDelete(MouseEvent e) {
		if (e == null)
			return false;
		int onMask = MouseEvent.BUTTON3_DOWN_MASK
				| MouseEvent.BUTTON2_DOWN_MASK;
		return (e.getModifiersEx() & onMask) == onMask;
	}

	public static boolean isGetAttributes(MouseEvent e) {
		if (e == null)
			return false;
		int onMask = MouseEvent.BUTTON3_DOWN_MASK
				| MouseEvent.BUTTON1_DOWN_MASK;
		return (e.getModifiersEx() & onMask) == onMask;
	}

	public static boolean isTwoClickNoOp(MouseEvent e) {
		if (e == null)
			return false;
		int onMask = MouseEvent.BUTTON2_DOWN_MASK
				| MouseEvent.BUTTON1_DOWN_MASK;
		return (e.getModifiersEx() & onMask) == onMask;
	}

	public static boolean wasDeleteClicked() {
		return isDelete(_lastMouseClick);
	}
}
