package org.expeditee.io;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameIO;
import org.expeditee.items.Item;

public abstract class DefaultFrameWriter extends ItemWriter implements
		FrameWriter {

	protected String _filename = null;

	protected String _output = null;

	protected ProxyWriter _writer = null;

	protected String _format = "";

	protected boolean _running = true;

	protected boolean _stop = false;

	public void setOutputLocation(String filename) {
		_filename = filename;
	}

	public String writeFrame(Frame toWrite) throws IOException {
		try {
			initialise(toWrite);

			outputFrame(toWrite);

			_running = false;
			return finaliseFrame();
		} catch (IOException ioe) {
			_running = false;
			throw ioe;
		}

	}

	protected void writeStartFrame(Frame starting) throws IOException {
		if (starting.getTitle() != null) {
			if (starting.getTitle().isAnnotation())
				this.writeAnnotationTitle(starting.getTitle());
			else
				this.writeTitle(starting.getTitle(), starting.getItems());
		}
	}

	protected void writeEndFrame(Frame ending) throws IOException {
	}

	protected void initialise(Frame start) throws IOException {
		if (_filename == null)
			_filename = FrameIO.EXPORTS_DIR + start.getFrameName() + _format;

		if (_filename.toLowerCase().equals("clipboard")) {
			_writer = new ProxyWriter(true);
			_output = "Clipboard";
		} else {
			if (_filename.contains(File.separator)) {
				String extTest = _filename.substring(_filename
						.lastIndexOf(File.separator) + 1);
				if (!extTest.contains("."))
					_filename += _format;
			} else if (!_filename.contains("."))
				_filename += _format;

			if (!_filename.contains(File.separator))
				_filename = FrameIO.EXPORTS_DIR + _filename;

			File test = new File(_filename);

			test = test.getParentFile();
			if (test != null && !test.exists()) {
				test.mkdirs();
			}

			_writer = new ProxyWriter(new FileWriter(_filename));
			_output = _filename;
		}
	}

	protected String finalise() throws IOException {
		try {
			_writer.flush();
			_writer.close();
		} catch (IOException ioe) {
		} finally {
			_writer.close();
		}

		return " exported to " + _output;
	}

	protected String finaliseFrame() throws IOException {
		return "Frame" + finalise();
	}

	protected void outputFrame(Frame toWrite) throws IOException {
		writeStartFrame(toWrite);

		for (Item i : toWrite.getItems()) {
			if (_stop)
				return;
			writeItem(i);
		}

		writeEndFrame(toWrite);
	}

	public boolean isRunning() {
		return _running;
	}

	public void stop() {
		_stop = true;
	}

	public String getFileContents() {
		return "Not Supported";
	}
}
