package org.expeditee.stats;

import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.sql.Time;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.expeditee.gui.DisplayIO;
import org.expeditee.gui.FrameUtils;
import org.expeditee.gui.TimeKeeper;
import org.expeditee.io.Logger;

public class SessionStats {

	private static final long MILLISECONDS_PER_MINUTE = 6000;

	private static Date _FrameAccessTime = new Date();

	private static List<String> _FrameEvents = new ArrayList<String>();

	// statistics counters
	private static final int MOUSE_BUTTONS = 3;

	private static StringBuffer _FramesEdited = new StringBuffer();

	private static int[] _MouseCounters = new int[MOUSE_BUTTONS];

	private static int _AccessedFrames = -1;

	private static int _SavedFrames = 0;

	// the number of frames created via TDFC
	private static int _CreatedFrames = 0;

	// character counts
	private static int[] _CharCounts = new int[255];

	// how many items have been created thus far
	private static int _CreatedItems = 0;

	private static Date _StartTime = new Date();
	
	private static long _LastEvent = new Date().getTime();
	
	private static Time _DarkTime = new Time(0);
	
	private static final long DARK_TIME_THRESHOLD = 60000;

	private static int _EscapeCount = 0;

	private static boolean _StatsEnabled = true;

	public static String getCurrentStats() {
		StringBuffer stats = new StringBuffer("Date: ");

		stats.append(Logger.EasyDateFormat("ddMMMyyyy")).append("\n");
		stats.append("SessionTime: ").append(getTimeElapsed()).append("\n");
		stats.append("DarkTime: ").append( new SimpleDateFormat("m").format(_DarkTime) ).append("\n");
		
		stats.append(getFrameStats());
		stats.append(getResponseStats()).append("\n");

		stats.append(getCharacterStats());

		stats.append("Version: ").append(DisplayIO.TITLE);

		return stats.toString();
	}

	private static String getTimeElapsed() {
		SimpleDateFormat formatter = new SimpleDateFormat("HH:mm");
		Date currentTime = new Date();
		Time elapsedTime = new Time(currentTime.getTime()-_StartTime.getTime());
		String time = formatter.format(_StartTime);
		time += " --> " + formatter.format(currentTime);
		time += " = " + (new SimpleDateFormat("m").format(elapsedTime));

		return time;
	}

	private static String getResponseStats() {
		String stats = "ResponseTime: "
				+ TimeKeeper.Formatter.format(FrameUtils.getLastResponseTime())
				+ "sec, ";

		if (_AccessedFrames > 0)
			stats += TimeKeeper.Formatter.format(FrameUtils
					.getResponseTimeTotal()
					/ _AccessedFrames)
					+ "avg";
		else
			stats += "--avg";

		return stats;
	}

	private static String getFrameStats() {
		return getFrameStats(true);
	}

	private static String getFrameStats(boolean newline) {
		StringBuffer stats = new StringBuffer();
		appendStat(stats, "FramesAccessed", _AccessedFrames, newline);
		appendStat(stats, "FramesEdited", _SavedFrames, newline);
		return stats.toString();
	}

	private static int getMinutesUsed() {
		long elapsedTime = new Date().getTime() - _StartTime.getTime() - _DarkTime.getTime();
		
		return (int) (elapsedTime / MILLISECONDS_PER_MINUTE);
	}

	private static String getCharacterStats() {
		int[] counts = _CharCounts;

		int chars = 0;

		for (int i = 'A'; i <= 'Z'; i++)
			chars += counts[i];

		for (int i = 'a'; i <= 'z'; i++)
			chars += counts[i];

		for (int i = '0'; i <= '9'; i++)
			chars += counts[i];

		chars -= counts[KeyEvent.VK_BACK_SPACE];
		chars -= counts[KeyEvent.VK_DELETE];

		int EOS = counts['.'] + counts[','] + counts['!'] + counts['?'];

		int punct = counts[' '] + counts[';'] + counts[':'];
		chars += counts['('] + counts[')'] + counts['\''] + counts['"']
				+ counts['+'] + counts['='] + counts['-'];

		StringBuffer stats = new StringBuffer();
		appendStat(stats, "Chars", chars + punct + EOS);
		appendStat(stats, "Words", punct + EOS);
		appendStat(stats, "Sentences", EOS);
		appendStat(stats, "Items", _CreatedItems);
		appendStat(stats, "Frames", _CreatedFrames);
		appendStat(stats, "Escape", _EscapeCount);
		appendStat(stats, "Left", _MouseCounters[MouseEvent.BUTTON1 - 1]);
		appendStat(stats, "Middle", _MouseCounters[MouseEvent.BUTTON2 - 1]);
		appendStat(stats, "Right", _MouseCounters[MouseEvent.BUTTON3 - 1]);

		return stats.toString();
	}

	/**
	 * Appends a single stats to a string buffer containing a collection of
	 * stats.
	 * 
	 * @param stats
	 *            The string buffer to append the stat onto
	 * @param name
	 *            The name of the stat
	 * @param value
	 *            The new value for the stat
	 */
	private static void appendStat(StringBuffer stats, String name, int value,
			boolean newline) {
		// prevent divide by zero errors
		if (value < 0)
			value = 0;

		int perHour = value * 60 / Math.max(1, getMinutesUsed());

		stats.append(value).append(" ").append(name).append(" @ ").append(perHour).append("/hour");
		if (newline)
			stats.append("\n");
		else
			stats.append(" ");

	}

	private static void appendStat(StringBuffer stats, String name, int value) {
		appendStat(stats, name, value, true);
	}

	public static void resetStats() {
		_StartTime = new Date();
		_AccessedFrames = 0;
		_SavedFrames = 0;
		_CreatedFrames = 0;
		_CharCounts = new int[255];
		_CreatedItems = 0;
		_EscapeCount = 0;
		_DarkTime.setTime(0);
		_MouseCounters = new int[MOUSE_BUTTONS];
	}

	/**
	 * Called signal that a frame has been accessed.
	 * 
	 */
	public static void AccessedFrame() {
		if (_StatsEnabled) {
			_AccessedFrames++;
			_FrameEvents.clear();
			_FrameAccessTime = new Date();
		}
	}

	public static void SavedFrame(String frameName) {
		FrameEdited(frameName);
		_SavedFrames++;
	}

	public static void Escape() {
		_EscapeCount++;
	}

	public static void CreatedFrame() {
		_CreatedFrames++;
	}

	public static void CreatedItem() {
		_CreatedItems++;
	}

	/**
	 * Increments the count for a character that is typed.
	 * 
	 * @param ch
	 *            ascii value for the typed character
	 */
	public static void TypedChar(int ch) {
		UserEvent();
		_CharCounts[ch]++;
	}
	
	private static void UserEvent() {
		long thisEvent = new Date().getTime();
		long elapsedTime = thisEvent - _LastEvent;
		if (elapsedTime > DARK_TIME_THRESHOLD)
			_DarkTime.setTime(_DarkTime.getTime() + elapsedTime);
		_LastEvent = thisEvent;
	}

	public static void AddFrameEvent(String description) {
		Date elapsedTime = new Date((new Date()).getTime()
				- _FrameAccessTime.getTime());

		_FrameEvents.add(Logger.EasyDateFormat("mm:ss", elapsedTime) + " "
				+ DisplayIO.getMouseX() + " " + DisplayIO.getMouseY() + " "
				+ description);
	}

	public static String getFrameEventList() {
		StringBuilder eventList = new StringBuilder();
		for (String s : _FrameEvents)
			eventList.append(s + '\n');

		return eventList.toString();
	}

	public static void MouseClicked(int button) {
		UserEvent();
		_MouseCounters[button - 1]++;
	}

	public static void setEnabled(boolean value) {
		_StatsEnabled = true;
	}

	private static void FrameEdited(String name) {
		_FramesEdited.append(Logger.EasyDateFormat("ddMMMyyyy-HHmm:ss"))
				.append("[").append(name).append("]\n");
	}

	public static String getFramesEdited() {
		return _FramesEdited.toString();
	}
}
